/*------------------------------------------------------------------*/
/*  YHF_Tone.cpp                                                    */
/*                                                                  */
/* DL4YHF's tiny tone signal generator.                             */
/*                                                                  */
/*       Originally just a simple routine to generate tones         */
/*       with the PC's speaker, now adapted for Windoze+Soundcard.  */
/*                                                                  */
/* Copyright (c) 2001 by Wolfgang Buescher                          */
/*                                                                  */
/* Autor: Wolfgang Buescher, (www.qsl.net/DL4YHF)                   */
/*                                                                  */
/* Revision history:                                                */
/*  10.Sep.2001 :                                                   */
/*        Recycled from another project.                            */
/*                                                                  */
/*------------------------------------------------------------------*/

#ifndef _YHF_TONE_  // avoid multiple inclusion
 #define _YHF_TONE_ 1


//------------ External References (from "somewhere" :-) --------------------
#ifdef __WIN32__
#ifdef __cplusplus
extern "C" {
#endif
/* Routines for direct hardware access.
 * Usually located somewhere in the code of the main module.
 * Almost certain that this won't work under Windows NT / 2000 .
 */
extern void CPU_outp(WORD portid, BYTE data);
extern BYTE CPU_inp(WORD portid);
#ifdef __cplusplus
           }
#endif
#endif /* __WIN32__ */



//------------ Defines ------------------------------------------------------
#define YHF_TONE_SOFTWARE_VERSION  1

// Valid "audio output devices", used for the iAudioDevice parameter:
#define YHF_TONE_AUDIO_DEVICE_NONE 0
#define YHF_TONE_AUDIO_DEVICE_SPKR 1 /* this is the INTERNAL "PC-Speaker"   */
#define YHF_TONE_AUDIO_DEVICE_MIDI 2 /* this is the midi-driven Synthesizer */


// Music Instruments, accurding the General MIDI "Program Number",
//       see http://de.wikipedia.org/wiki/General_MIDI .
//  Here are just SOME of them:
#define YHF_INSTRUMENT_PIANO   1
#define YHF_INSTRUMENT_ORGAN   19 /* "Rock Organ", the others sound terrible */
#define YHF_INSTRUMENT_TRUMPET 57
#define YHF_INSTRUMENT_PICCOLO 73
#define YHF_INSTRUMENT_FLUTE   74
#define YHF_INSTRUMENT_SQUARE  81
#define YHF_INSTRUMENT_SAWTOOTH 82



//------------ Data Types ---------------------------------------------------


//------------ Variables ----------------------------------------------------


//------------ Prototypes ---------------------------------------------------


//---------------------------------------------------------------------------
int YHF_InitTones(void);   // Must be called once to initialize the sound output.
void YHF_CloseTones(void); // Must be called once on exit, to close handles etc.


//---------------------------------------------------------------------------
void YHF_GenerateTone(int iAudioDevice, int iFrequencyHz);
  /* Starts the output of a tone with the PC speaker, or the MIDI synth .
   * WARNING: CALLS A NASTY PORT ACCESS ROUTINE IN THE MAIN FORM,
   *          AND MUST NOT BE CALLED FROM ANY THREAD !!!
   *          THIS IS DEFINITELY A WAY HOW IT SHALL NOT BE DONE
   *          IN A WELL-BEHAVING WINDOZE PROGRAM.
   */





#endif // _YHF_TONE_  to avoid multiple inclusion

// EOF <YHF_Tone.h>